/*
 * intcom.c
 *     interactive communication program between the external
 *     controller and the instrument
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>

#include "decl-32.h"
#include "..\gpiblib.h"

VOID process ();
VOID inputline ();
VOID chkreplace ();
VOID ReadOutputbuf ();
VOID setarg ();
VOID bnull ();
char *getfile ();
int viewfile ();
int execfile ();
VOID statusbyte ();
VOID helpmessg ();
VOID resetes ();
int chkstatus ();
VOID redirect ();

# define    MAX_BUF     (128)
# define    MAX_ARG     (10)
# define    RD0         (0)
# define    RD1         (1)
# define    RD2         (2)
# define    RD3         (3)
# define    ON          (1)
# define    OFF         (-1)

extern int  GpibStatus;

char    *getfile ();

int     argCnt;
jmp_buf env;
char    *argVal [MAX_ARG];
char    readbuf [MAX_BUF + 1];
char    replace [MAX_BUF + 1];
char    assmble [MAX_BUF + 1];
char    stack [MAX_BUF + 1];

char    openDevice [MAX_BUF + 1];

FILE    *ifd;       /* Input file descriptor */
FILE    *ofd;       /* Output file descriptor */
char    *rfile;
int     rflag;      /* > : RD1, >> : RD2,  < : RD3, none : RD0  */
int     sflag;

VOID
main (argc, argv)
    int     argc;
    char    *argv[];
{
    printf ("\nintcom  -  interactive GPIB communication program\n");
    printf ("Copyright (C) Tektronix Japan, Ltd. All rights reserved.\n");

    if (argc == 1)
        strcpy (openDevice, "dev1");      /* default device */
    else if (argc == 2)
        strcpy (openDevice, argv[1]);     /* specified device */
    else
    {
        printf ("\nUsage: intcom [<device>]\n");
        printf ("  where <device> is a name set by 'ibconf'");
        printf (" such as 'dev1' (default).\n");
        exit (1);
    }

    printf ("  Type 'help' to display help messages, 'q' to exit.\n\n");

    GpibOpen (openDevice);      /* find specified device */
    process ();                 /* communication process */
    GpibClose ();
}

/*
 * Communication process.
 */
VOID
process ()
{
    /* initialize */

    stack[0] = '\0';
    replace[0] = '\0';
    ofd = stdout;       /* Output to standard output */
    ifd = stdin;        /* Input from standard input */
    resetes (1);        /* Set enable registers */

    /* Start interactive communication */

    for ( ; ; )
    {
        /* check status byte and dequeue stacked events */
        chkstatus ();

        /* reset output stream to stdout */
        if (ofd != stdout)
        {
            fclose (ofd);
            ofd = stdout;
        }

        /* get and parse command line */
        inputline ();

        /* terminate interactive process */
        if (strcmp (argVal[0], "q") == 0
            || strcmp (argVal[0], "quit") == 0)
            break;

        /* redirect */
        if (rflag == RD1 || rflag == RD2)
            redirect ();

        /* execute built-in command */
        if (strcmp (argVal[0], "exec") == 0)
            execfile (argCnt, argVal[1]);
        else if (strcmp (argVal[0], "view") == 0)
            viewfile (argCnt, argVal[1]);
        else if (strcmp (argVal[0], "help") == 0)
            helpmessg (argCnt);
        else if (strcmp (argVal[0], "status") == 0)
            statusbyte (argCnt);
        else if (strcmp (argVal[0], "resetes") == 0)
            resetes (argCnt);

        /* Transmit GPIB command. If query, read output queue immediately */
        else if (argCnt > 0)
        {
            if (rflag == RD3)
            {
                fprintf (stderr, "syntax error: (%s)\n", assmble);
            }
            else if (strcmp (assmble, "?") == 0)
            {
                ReadOutputbuf ();  
            }
            else
            {
                if (GpibWriteString (assmble) < 0)
                {
                    GpibError ("Write Error:");
                    continue;
                }
                if (strchr (assmble, '?'))
                    ReadOutputbuf ();
            }
        }
        else if (rflag == RD3)
        {
            GpibWriteFile (rfile);
        }
    }
}

/*
 * Parse input command line and store results
 *
 *  *argVal[] - command and arguments delimited by space
 *  argCnt    - number of arguments
 *  assemble  - concatenation of *argVal[]
 *  rflag     - '>', '>>', '<', or none
 *  rfile     - input or output file to be redirected
 *
 * If the string '!!' exists in command line, it is replaced with
 * previous command line.
 */
VOID
inputline ()
{
    int     i;      /* loop index */

    for ( ; ; )
    {
        printf ("%s>> ", openDevice);
        if (fgets (readbuf, MAX_BUF, ifd) == NULL) /* Read one line */
        {
            if (ifd == stdin)
            {
                fprintf (stderr, "Detected system error: restart the program\n");
                exit (1);
            }
            fclose (ifd);
            ifd = stdin;
            continue;
        }
        if (ifd != stdin)
            printf ("%s", readbuf);
        chkreplace (readbuf);
        setarg (replace);            /* Parse input line */
        if (!(argCnt < 1 && rflag == RD0))    /* Check input */
            break;
    }

    assmble[0] = '\0';
    for (i = 0; i < argCnt; i++)          /* Assemble line    */
    {
        strcat (assmble, argVal[i]);
        if (i + 1 < argCnt)
            strcat (assmble, " ");
    }
    strcpy (stack, assmble);
    if (rflag != RD0)
    {
        strcat (stack, " ");
        strcat (stack, (rflag == RD1) ? ">" : ((rflag == RD2) ? ">>" : "<") );
        strcat (stack, " ");
        strcat (stack, rfile);
    }
}

/*
 * Check for '!!' and replace if existed.
 */
VOID
chkreplace (s)
    char    *s;
{
    char    *p = stack;
    char    *r = replace;
    int     cc = 0;     /* flag: replaced or not */

    /* Replace '!!' with previous input line */

    while (*s)
    {
        if (strncmp (s, "!!", 2) == 0)
        {
            for (p = stack; *p; )
                *r++ = *p++;
            s += 2;
            p = stack;
            cc++;
        }
        else
            *r++ = *s++;
    }
    *r = '\0';

    if (cc != 0)
        printf ("%s", replace);
}

/*
 * Read from output queue, and write stdin or file.
 */
VOID
ReadOutputbuf ()
{
    char    *p;
    char    sc;         /* Store one character */
    int     i;          /* Loop index */

    /* take out */
    if (rflag == RD1 || rflag == RD2)
    {
        for ( ; ; )
        {
                if (GpibRead (readbuf, MAX_BUF) < 0)
            {
                GpibError ("Read Error:");
                break;
            }
            for (p = readbuf, i = 0; i < ibcnt; i++)
                putc (*p++, ofd);
            if (GpibStatus & (ERR | TIMO | END))
                break;
        }
    }
    else        /* Read the Output Queue, and write to stdout   */
    {
        sc = ' ';
        for ( ; ; )
        {
                if (GpibRead (readbuf, MAX_BUF) < 0)
            {
                GpibError ("Read Error:");
                break;
            }
            for (p = readbuf, i = 0; i < ibcnt; i++, p++)
            {
                if (*p == ';')
                {
                    sc = *p;
                }
                else
                {
                    if (*p == ':' && sc == ';')
                        putc ('\n', ofd);
                    else if (sc == ';')
                        putc (sc, ofd);
                    putc (*p, ofd);
                    sc = *p;
                }
            }
            if (GpibStatus & (ERR | TIMO | END))
                break;
        }
    }
}

/*
 * Parse command line.
 */
VOID
setarg (str)
    char    *str;
{
    char    *s = str;
    char    *p = str;

    sflag = OFF;
    rflag = RD0;
    argCnt = 0;

    for ( ; *s; s++)
    {
        switch ((int)*s)
        {
        case ' ':
            sflag = OFF;
            bnull(p, s);
            break;

        case '\n':
            sflag = OFF;
            bnull (p, s);
            --s;
            break;

        case '>':
            bnull (p, s);
            if ( *(s+1) == '>')
            {
                rflag = RD2;
                *++s = '\0';
            }
            else
            {
                rflag = RD1;
            }       
            sflag = OFF;
            s = getfile (++s);
            s--;
            break;

        case '<' :
            bnull (p, s);
            if ( *(s+1) == '<')
                *++s = '\0';
            rflag = RD3;
            sflag = OFF;
            s = getfile (++s);
            s--;
            break;

        default :
            if (sflag == OFF)
            {
                sflag = ON;
                argVal[argCnt] = p = s;
                ++argCnt;
            }
        }
    }
}

/*
 * Put '\0' at the end of the command and arguments.
 */
VOID
bnull (p, s)
    char    *p, *s;
{
    *s-- = '\0';
    while (p < s && *s == ' ')
        *s-- = '\0';
}

/*
 * Extract file name after '<', '>', or '>>'
 */

char *getfile (s)
    char    *s;
{
    for ( ; *s == ' '; s++)
        ;
    rfile = s;
    for (;*s && *s != '\n' && *s != ' ' && *s != '>' && *s != '<'; s++)
        ;
    *s = '\0';
    return ++s;
}

/*
 * 'view' built-in command.
 */
int
viewfile (n, name)
    int     n;
    char    *name;
{
    FILE    *tfd;
    int     c;

    if (n != 2)
    {
        fprintf (stderr, "usage: view file-name\n");
        return 0;
    }

    if ((tfd = fopen (name, "r")) == NULL)
    {
        fprintf(stderr, "can't open file (%s)\n", name);
        return 0;
    }

    while ((c = getc (tfd)) != EOF)
        putc (c, ofd);

    fclose (tfd);
    return 1;
}

/*
 * 'exec' built-in command.
 */
int
execfile (n, name)
    int     n;
    char    *name;
{
    FILE    *tfd;

    if (n != 2)
    {
        fprintf (stderr, "usage: exec file-name\n");
        return 0;
    }

    if ((tfd = fopen(name, "r")) == NULL)
    {
        fprintf(stderr, "can't open file (%s)\n", name);
        return 0;
    }
    ifd = tfd;
    return 1;
}

/*
 * 'status' built-in command.
 */
VOID
statusbyte (n)
    int     n;
{
    if (n != 1)
        fprintf (stderr, "Arguments are ignored.\n\n");

    fprintf (ofd, "Status byte: (%X)h\n", GpibSerialPoll ());
}

static char hmessg[] = {"\n\
You may use the following commands.\n\n\
[1] GPIB commands: All commands and queries are available. The query\n\
   response is immediately output to the standard output.\n\n\
[2] Built-in commands:\n\
   help - displays this message\n\
   view filename - displays the contents of the file\n\
   exec filename - reads command lines from the file until EOF\n\
   status - reads the status byte from the instrument\n\
   resetes - resets the event/status reporting system to the default\n\
   quit - terminates execution of this program\n\
   !! - repeats the previous command line\n\n\
[3] Redirection: The standard input/output may be redirected.\n\
   < filename - The file is used as standard input.\n\
   > filename - The file is used as standard output.\n\
   >> filename - The file is used as standard output in append mode.\n\n\
"};

/*
 * 'help' built-in command.
 */
VOID
helpmessg (n)
    int     n;
{
    char    *p = hmessg;
 
    if (n != 1)
    {
        fprintf (stderr, "Arguments are ignored.\n\n");
    }
    while (*p)
        putc (*p++, ofd);
}

/*
 * 'resetes' built-in command.
 */
VOID
resetes (n)
    int     n;
{
    if (n != 1)
        fprintf (stderr, "Arguments are ignored.\n\n");

    if (GpibWriteString ("*CLS;*ESE 58;*SRE 48") < 0)
        GpibError ("Write Error: can't set enable registers");
}

/*
 * Read event queue if event is being stacked.
 *
 * '\n' is added after event code and event message.
 */
int
chkstatus ()
{
    int     ccount = 0;
    int     i;
    char    *p;

    /* check for the ESB bit in the SBR */
    if (! (GpibSerialPoll () & 0x20))
    {
        GpibTimeOut (T1s);      /* Wait for further 1s */
        GpibWait (TIMO | SRQI | RQS | END);
        GpibTimeOut (T3s);      /* Reset to 3s */
        if (! (GpibSerialPoll () & 0x20))
            return 0;
    }

    /* prepare to take out */
    if (GpibWriteString ("*ESR?") < 0
        || GpibWriteString (":SYSTEM:ERROR?") < 0)
    {
        GpibError ("Write Error:");
        return 0;
    }
    
    /* read the event queue */
    for ( ; ; )
    {
        if (GpibRead (readbuf, MAX_BUF) < 0)
        {
            GpibError ("Read Error:");
            return 0;
        }
        for (p = readbuf, i = 0; i < ibcnt; i++, p++)
        {
            putc (*p, ofd);
            if (*p == ',' && ccount == 1)
            {
                putc('\n', ofd);
                ccount = 0;
            }
            else if (*p == ',')
                ccount++;
        }
        if (GpibStatus & (ERR | TIMO | END))
            return 0;
    }

    if (GpibWriteString ("HEADER ON") < 0)
    {
        GpibError ("Write Error:");
        return 0;
    }
    return 1;
}

/*
 * Open file for output.
 */
VOID
redirect ()
{
    FILE    *tfd;       /* tempolary file descriptor */

    if (rflag == RD1)
        tfd = fopen (rfile, "wb");
    else /* RD2 */
        tfd = fopen (rfile, "ab");

    if (tfd == NULL)
        fprintf (stderr, "can't open file (%s)\n", rfile);
    else
        ofd = tfd;
}
