/*
 *   putequ.c
 *
 *   a sample program to transfer the contents of the equation file
 *   to the data storage of AWG.
 *
 *   Note: The return status of the GPIB library function is not
 *         fully checked in order to keep the code simple and
 *         understandable. You would be able to improve it
 *         with AWG local operations.
 *
 *   Copyright (C) Tektronix Japan, Ltd. All rights reserved.
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "decl-32.h"
#include "..\gpiblib.h"

VOID ShowTitle();
VOID ShowUsage();
VOID MakeFilePathName();
int ReadEquDataSize();
int ReadEquData();
VOID SendEquData();

#define TRUE    1
#define FALSE   0

long    SendLine; /* line of the equation data */
fpos_t	SendSize;
char    buffer[512];
char    FilePath[256];
char    FileName[41];



/*
 * the main routine
 */
VOID
main(argc, argv)
    int     argc;
    char    *argv[];
{
    FILE  *fp;		/* input equation data file */
    char *equData;	/* pointer to allocated memory */

    /* opening message */

    ShowTitle();

    /* check command line argument */

    if (argc != 3)
    {
        ShowUsage();
        exit(1);
    }

	MakeFilePathName(argv[2]);

    /* read header parameters of the word-oriented pattern file */

    if ((fp = fopen(argv[1], "rb")) == NULL)
    {
        fprintf(stderr, "Error: Unable to open file \"%s\"\n", argv[1]);
        ShowUsage();
        exit(1);
    }

	if (!ReadEquDataSize(fp))
	{
		fprintf(stderr, "Error: Unable to read file \"%s\"\n", argv[1]);
		exit(1);
	}

    fprintf(stderr, "sending data from file \"%s\"\n", argv[1]);
	fprintf(stderr, "             to   file \"%s%s\"\n", FilePath, FileName);

    /* initialize GPIB communication */

    GpibOpen("dev1");

    /* allocate memory to hold word-oriented pattern block */

    equData = malloc((size_t)(SendSize*2L));
    if (equData == NULL)
    {
        fprintf(stderr, "Error: Unable to allocate work memory\n");
        GpibClose();
        exit(1);
    }

    /* read and send the equation data block using work memory */

	fprintf(stderr, "Size=%ld\n", (int)SendSize);
    fprintf(stderr, "sending equation data\n");
    ReadEquData(fp, equData);
    SendEquData(equData);

    GpibCheckError();
    /* close file, memory and GPIB communication */

    fclose(fp);
    free(equData);
    GpibClose();
}

/*
 * show opening message
 */
VOID
ShowTitle()
{
    fprintf(stderr, "putequ - send the Waveform equation data\n");
    fprintf(stderr, "  Copyright (C) Tektronix Japan, Ltd. All rights reserved.\n\n");
}

/*
 * show command syntax
 */
VOID
ShowUsage()
{
    fprintf(stderr, "Usage: putequ <aource-file> <destination-file>\n");
    fprintf(stderr, "  where <source-file> is the name of the equation data file.\n");
    fprintf(stderr, "  where <destination-file> is the name of the AWG's mass storage file.\n");
}

/*
 * make a file path & filename from argument
 */
VOID
MakeFilePathName(file)
        char *file;
{
        char *p;

        p = strrchr(file, '/');
        if (p == NULL) {
                strcpy(FilePath, "./");
                strcpy(FileName, file);
        } else {
                strncpy(FilePath, file, strlen(file)-strlen(p)+1);
                strcpy(FileName, p+1);
        }
}

/*
 * read data file size
 */
int
ReadEquDataSize(fp)
    FILE  *fp;
{
	if (!fseek(fp, 0L, SEEK_END)) {
		if (!fgetpos(fp, &SendSize)) {
			if (!fseek(fp, 0L, SEEK_SET)) {
				return TRUE;
			}
		}
	}
	return FALSE;
}

/*
 * read data block from the word-oriented pattern file
 */
int
ReadEquData(fp, equData)
    FILE *fp;
    char *equData;
{
	char *p;

	SendLine = 0L;
    p = equData;
	/* read word data block */
    while (fgets(buffer, 128, fp) != NULL)
    {
        buffer[128] = '\0';
		strcpy (p, buffer);
		p += strlen(buffer);
		SendLine++;
    }
	return TRUE;
}

/*
 * send the word-oriented pattern data
 */
VOID
SendEquData(equData)
    char *equData; /*  */
{
    char work[16];

    /* send command part and block header */

    GpibWriteString(":MMEMORY:MSIS \"MAIN\""); /* MassStorage is the Internal HD */
    sprintf(buffer, ":MMEMORY:CDIRECTORY \"%s\"", FilePath);
    GpibWriteString(buffer);

	sprintf(work, "%d", strlen(equData));
    sprintf(buffer, ":MMEMORY:DATA \"%s\",#%d%s", FileName, strlen(work), work);
    GpibControlEOI(0); /* disable sending EOI */
    GpibWriteString(buffer); /* continued to write below */
    
	GpibControlEOI(1); /* enable sending EOI */
	GpibWriteString(equData); /* send equation data */
}
